import collections
import functools
import numpy as np
from .vector import arggroupby as _arggroupby, vector

try:
    from future_builtins import zip
    from collections import Mapping  # pragma: no cover
except ImportError:
    from typing import Mapping


def arggroupby(keys):
    """Generate unique keys with corresponding index arrays."""
    keys = np.asarray(keys)
    order = np.argsort(keys)
    keys, counts = np.unique(keys[order], return_counts=True)
    start = 0
    for key, stop in zip(keys, counts.cumsum()):
        yield key, order[start:stop]
        start = stop


def groupby(keys, *arrays):
    """Generate unique keys with associated groups."""
    arrays = list(map(np.asarray, arrays))
    try:
        items = _arggroupby(np.asarray(keys).astype(np.intp, casting='safe', copy=False))
    except TypeError:  # fallback to sorting
        items = arggroupby(keys)
    for key, values in items:
        yield (key,) + tuple(arr[values] for arr in arrays)


class matrix(collections.defaultdict):
    """A sparse vector of vectors."""

    def __init__(self, data=(), copy=True):
        super(matrix, self).__init__(vector)
        (self if copy else super(matrix, self)).update(data)

    @classmethod
    def cast(cls, data):
        return cls(data, copy=False)

    @property
    def row(self):
        """COO format row index array of the matrix"""
        return np.concatenate([np.full(len(self[key]), key) for key in self])

    @property
    def col(self):
        """COO format column index array of the matrix"""
        return np.concatenate([vec.keys() for vec in self.values()])

    @property
    def data(self):
        """COO format data array of the matrix"""
        return np.concatenate([vec.values() for vec in self.values()])

    def update(self, data):
        """Update from mapping or iterable."""
        if isinstance(data, Mapping):
            for key in data:
                self[key].update(data[key])
        else:
            for key, value in data:
                self[key].update(value)

    def __iadd__(self, other):
        if isinstance(other, Mapping):
            for key in other:
                self[key] += other[key]
        else:
            self.map(vector.__iadd__, other)
        return self

    def __add__(self, other):
        return type(self)(self).__iadd__(other)

    def __imul__(self, other):
        if isinstance(other, Mapping):
            for key in set(self).difference(other):
                del self[key]
            for key in self:
                self[key] *= other[key]
        else:
            self.map(vector.__imul__, other)
        return self

    def __mul__(self, other):
        if isinstance(other, Mapping):
            return self.cast((key, self[key] * other[key]) for key in set(self).intersection(other))
        return self.map(vector.__mul__, other)

    def sum(self, axis=None):
        """Return sum of matrix elements across axis, by default both."""
        if axis in (0, -2):
            return functools.reduce(vector.__iadd__, self.values(), vector())
        if axis in (1, -1):
            return self.map(np.sum)
        if axis is None:
            return sum(map(np.sum, self.values()))
        raise np.AxisError("axis {} is out of bounds".format(axis))

    def map(self, func, *args, **kwargs):
        """Return matrix with function applies across vectors."""
        result = {key: func(self[key], *args, **kwargs) for key in self}
        return self.cast(result) if all(isinstance(value, vector) for value in result.values()) else result

    def filter(self, func, *args, **kwargs):
        """Return matrix with function applies across vectors."""
        return self.cast((key, vec) for key, vec in self.items() if func(vec, *args, **kwargs))

    @classmethod
    def fromcoo(cls, row, col, data):
        """Return matrix from COOrdinate format arrays."""
        return cls.cast((key, vector(col, data)) for key, col, data in groupby(row, col, data))

    def transpose(self):
        """Return matrix with reversed dimensions."""
        return self.fromcoo(self.col, self.row, self.data)

    T = property(transpose)

    def __matmul__(self, other):
        other = other.transpose()
        return self.cast((key, vector(other.map(self[key].dot))) for key in self)



import numpy as np
import pytest
from spector import matrix, vector


def test_basic():
    mat = matrix({'a': [0], 'b': [1]})
    assert np.array_equal(np.sort(mat.row), np.array(['a', 'b']))
    assert np.array_equal(np.sort(mat.col), np.array([0, 1]))
    assert np.array_equal(mat.data, np.array([1.0, 1.0]))
    mat.update([('b', [1]), ('c', [2])])
    assert dict(zip(zip(mat.row, mat.col), mat.data)) == {('a', 0): 1.0, ('b', 1): 2.0, ('c', 2): 1.0}


def test_math():
    mat = matrix({'a': [0], 'b': [1]})
    assert mat.sum() == 2.0
    mat += matrix({'b': [1], 'c': [2]})
    assert mat.sum(1) == {'a': 1.0, 'b': 2.0, 'c': 1.0}
    assert (mat + 1).sum(-1) == {'a': 2.0, 'b': 3.0, 'c': 2.0}
    vec = vector({0: 1.0, 1: 2.0, 2: 1.0})
    assert mat.sum(0).equal(vec)
    assert mat.sum(-2).equal(vec)
    with pytest.raises(ValueError):
        mat.sum(axis=2)

    mat = matrix({'a': [0], 'b': [1]})
    other = matrix({'b': {1: 2.0}, 'c': [2]})
    (key, vec), = (mat * other).items()
    assert key == 'b' and vec.equal(vector({1: 2.0}))
    mat *= other
    assert list(mat) == ['b'] and vec.equal(*mat.values())
    (key, vec), = (mat * 2).items()
    assert vec.values() == np.array([4.0])
    mat *= 2
    assert vec.equal(*mat.values())


def test_funcs():
    mat = matrix({'a': [0]})
    assert mat.map(np.sum, dtype=int) == {'a': 1}
    assert mat.filter(len) == mat
    assert not mat.filter(vector.dot, vector())
    with pytest.raises(TypeError):
        mat.transpose()
    assert not matrix.fromcoo([], [], [])
    mat = matrix.fromcoo([0, 0, 1, 1], [1, 2, 1, 2], [1, 2, 3, 4])
    assert mat.map(len) == {0: 2, 1: 2}
    assert mat.transpose().map(len) == mat.T.map(len) == {1: 2, 2: 2}
    mat = mat.__matmul__(mat.transpose())
    assert dict(mat[0]) == {0: 5.0, 1: 11.0}
    assert dict(mat[1]) == {0: 11.0, 1: 25.0}
    mat = matrix.fromcoo(list('abcba'), range(5), [1] * 5)
    assert mat.map(len) == {'a': 2, 'b': 2, 'c': 1}



from setuptools import setup, Extension

try:
    from Cython.Build import cythonize
except ImportError:
    cythonize = None

ext_module = Extension(
    'spector.vector',
    sources=['spector/vector' + ('.pyx' if cythonize else '.cpp')],
    extra_compile_args=['-std=c++11'],
    extra_link_args=['-std=c++11'],
)

setup(
    name='spector',
    version='0.2',
    description='Sparse vectors.',
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    author='Aric Coady',
    author_email='aric.coady@gmail.com',
    url='https://github.com/coady/spector',
    project_urls={'Documentation': 'https://spector.readthedocs.io'},
    license='Apache Software License',
    packages=['spector'],
    ext_modules=cythonize([ext_module]) if cythonize else [ext_module],
    install_requires=['numpy'],
    extras_require={'docs': ['m2r', 'nbsphinx', 'jupyter', 'pandas']},
    python_requires='>=2.7',
    tests_require=['pytest-cov'],
    keywords='sparse array vector matrix numpy scipy',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: Apache Software License',
        'Operating System :: OS Independent',
        'Programming Language :: Cython',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ],
)
